use crate::api::{publisher_client::PublisherClient, subscriber_client::SubscriberClient};
use crate::certs::get_pem;
use crate::error::Error;
use crate::subscription::Subscription;
use crate::topic::Topic;
use google_auth::TokenManager;
use once_cell::sync::OnceCell;
use tonic::transport::{Certificate, Channel, ClientTlsConfig};
use tonic::{IntoRequest, Request};

static DOMAIN_NAME: OnceCell<String> = OnceCell::new();
static ENDPOINT: OnceCell<String> = OnceCell::new();

#[derive(Clone)]
pub struct Client {
    pub(crate) project_id: String,
    pub(crate) publisher: PublisherClient<Channel>,
    pub(crate) subscriber: SubscriberClient<Channel>,
    pub(crate) token_manager: TokenManager,
}

impl Client {
    pub(crate) const SCOPES: [&'static str; 2] = [
        "https://www.googleapis.com/auth/cloud-platform",
        "https://www.googleapis.com/auth/pubsub",
    ];
    pub async fn new(project_id: &str) -> Result<Self, Error> {
        if let Ok(host) = std::env::var("PUBSUB_EMULATOR_HOST") {
            DOMAIN_NAME.set(host.clone()).unwrap();
            ENDPOINT.set(format!("http://{}", host)).unwrap();
        }
        let pem = get_pem();
        let tls_config = ClientTlsConfig::new()
            .ca_certificate(Certificate::from_pem(&pem))
            .domain_name(DOMAIN_NAME.get().unwrap());
        let channel = Channel::from_static(ENDPOINT.get().unwrap())
            .tls_config(tls_config)?
            .connect()
            .await?;
        let subscriber = SubscriberClient::new(channel.clone());
        let publisher = PublisherClient::new(channel);
        let token_manager = TokenManager::new(&Self::SCOPES).await?;
        Ok(Self {
            project_id: project_id.to_string(),
            subscriber,
            publisher,
            token_manager,
        })
    }
    pub(crate) async fn insert_authz_token<T: IntoRequest<T>>(
        &mut self,
        request: T,
    ) -> Result<Request<T>, Error> {
        let mut request = request.into_request();
        let token = self.token_manager.token().await?;
        let metadata = request.metadata_mut();
        metadata.insert("authorization", token.parse()?);
        Ok(request)
    }
    pub fn subscription(&self, subscription_name: &str) -> Subscription {
        let name = format!(
            "projects/{}/subscriptions/{}",
            &self.project_id, subscription_name
        );
        Subscription::new(self.clone(), name)
    }
    pub fn topic(&self, topic_name: &str) -> Topic {
        let name = format!("projects/{}/topics/{}", self.project_id, topic_name);
        Topic::new(self.clone(), name)
    }
}
